@@ -1,6 +1,6 @@
 // This file is a part of Julia. License is MIT: https://julialang.org/license
 
-/* Bring in definitions for `_OS_X_`, `PATH_MAX` and `PATHSEPSTRING`, `jl_ptls_t`, etc... */
+/* Bring in definitions for `_OS_X_`, `JL_PATH_MAX` and `PATHSEPSTRING`, `jl_ptls_t`, etc... */
 #include "../src/support/platform.h"
 #include "../src/support/dirpath.h"
 #include "../src/julia_fasttls.h"
@@ -25,9 +25,12 @@
 #include <stdint.h>
 
 #ifdef _OS_WINDOWS_
+
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
+
 #else
+
 #ifdef _OS_DARWIN_
 #include <mach-o/dyld.h>
 #endif
@@ -35,7 +38,6 @@
 #include <stddef.h>
 #include <sys/sysctl.h>
 #endif
-
 #define _GNU_SOURCE // Need this for `dladdr()`
 #include <stdio.h>
 #include <stdlib.h>
@@ -44,6 +46,7 @@
 #include <libgen.h>
 #include <unistd.h>
 #include <dlfcn.h>
+
 #endif
 
 // Borrow definition from `support/dtypes.h`
@@ -55,7 +58,7 @@
 # endif
 #define JL_HIDDEN
 #else
-# if defined(LIBRARY_EXPORTS) && defined(_OS_LINUX)
+# if defined(LIBRARY_EXPORTS) && defined(_OS_LINUX_)
 #  define JL_DLLEXPORT __attribute__ ((visibility("protected")))
 # else
 #  define JL_DLLEXPORT __attribute__ ((visibility("default")))
