@@ -1,4 +1,5 @@
 // This file is a part of Julia. License is MIT: https://julialang.org/license
+
 // This file defines an RPATH-style relative path loader for all platforms
 #include "loader.h"
 
@@ -15,7 +16,7 @@ extern "C" {
 #endif
 
 // Save DEP_LIBS to a variable that is explicitly sized for expansion
-static char dep_libs[512] = DEP_LIBS;
+static char dep_libs[1024] = DEP_LIBS;
 
 JL_DLLEXPORT void jl_loader_print_stderr(const char * msg)
 {
@@ -30,7 +31,7 @@ void jl_loader_print_stderr3(const char * msg1, const char * msg2, const char *
 }
 
 /* Wrapper around dlopen(), with extra relative pathing thrown in*/
-static void * load_library(const char * rel_path, const char * src_dir) {
+static void * load_library(const char * rel_path, const char * src_dir, int err) {
     void * handle = NULL;
 
     // See if a handle is already open to the basename
@@ -40,30 +41,33 @@ static void * load_library(const char * rel_path, const char * src_dir) {
             break;
     basename++;
 #if defined(_OS_WINDOWS_)
-    if ((handle = GetModuleHandleW(basename)))
+    if ((handle = GetModuleHandleA(basename)))
         return handle;
 #else
-    if ((handle = dlopen(basename, RTLD_NOLOAD | RTLD_NOW | RTLD_GLOBAL)))
+    // if err == 0 the library is optional, so don't allow global lookups to see it
+    if ((handle = dlopen(basename, RTLD_NOLOAD | RTLD_NOW | (err ? RTLD_GLOBAL : RTLD_LOCAL))))
         return handle;
 #endif
 
-    char path[2*PATH_MAX + 1] = {0};
+    char path[2*JL_PATH_MAX + 1] = {0};
     strncat(path, src_dir, sizeof(path) - 1);
     strncat(path, PATHSEPSTRING, sizeof(path) - 1);
     strncat(path, rel_path, sizeof(path) - 1);
 
 #if defined(_OS_WINDOWS_)
-    wchar_t wpath[2*PATH_MAX + 1] = {0};
-    if (!utf8_to_wchar(path, wpath, 2*PATH_MAX)) {
+    wchar_t wpath[2*JL_PATH_MAX + 1] = {0};
+    if (!utf8_to_wchar(path, wpath, 2*JL_PATH_MAX)) {
         jl_loader_print_stderr3("ERROR: Unable to convert path ", path, " to wide string!\n");
         exit(1);
     }
     handle = (void *)LoadLibraryExW(wpath, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
 #else
-    handle = dlopen(path, RTLD_NOW | RTLD_GLOBAL);
+    handle = dlopen(path, RTLD_NOW | (err ? RTLD_GLOBAL : RTLD_LOCAL));
 #endif
 
     if (handle == NULL) {
+        if (!err)
+            return NULL;
         jl_loader_print_stderr3("ERROR: Unable to load dependent library ", path, "\n");
 #if defined(_OS_WINDOWS_)
         LPWSTR wmsg = TEXT("");
@@ -94,7 +98,7 @@ static void * lookup_symbol(const void * lib_handle, const char * symbol_name) {
 }
 
 // Find the location of libjulia.
-char lib_dir[PATH_MAX];
+char lib_dir[JL_PATH_MAX];
 JL_DLLEXPORT const char * jl_get_libdir()
 {
     // Reuse the path if this is not the first call.
@@ -103,11 +107,11 @@ JL_DLLEXPORT const char * jl_get_libdir()
     }
 #if defined(_OS_WINDOWS_)
     // On Windows, we use GetModuleFileNameW
-    wchar_t libjulia_path[PATH_MAX];
+    wchar_t libjulia_path[JL_PATH_MAX];
     HMODULE libjulia = NULL;
 
     // Get a handle to libjulia.
-    if (!utf8_to_wchar(LIBJULIA_NAME, libjulia_path, PATH_MAX)) {
+    if (!utf8_to_wchar(LIBJULIA_NAME, libjulia_path, JL_PATH_MAX)) {
         jl_loader_print_stderr3("ERROR: Unable to convert path ", LIBJULIA_NAME, " to wide string!\n");
         exit(1);
     }
@@ -116,11 +120,11 @@ JL_DLLEXPORT const char * jl_get_libdir()
         jl_loader_print_stderr3("ERROR: Unable to load ", LIBJULIA_NAME, "!\n");
         exit(1);
     }
-    if (!GetModuleFileNameW(libjulia, libjulia_path, PATH_MAX)) {
+    if (!GetModuleFileNameW(libjulia, libjulia_path, JL_PATH_MAX)) {
         jl_loader_print_stderr("ERROR: GetModuleFileName() failed\n");
         exit(1);
     }
-    if (!wchar_to_utf8(libjulia_path, lib_dir, PATH_MAX)) {
+    if (!wchar_to_utf8(libjulia_path, lib_dir, JL_PATH_MAX)) {
         jl_loader_print_stderr("ERROR: Unable to convert julia path to UTF-8\n");
         exit(1);
     }
@@ -156,31 +160,78 @@ __attribute__((constructor)) void jl_load_libjulia_internal(void) {
     // Pre-load libraries that libjulia-internal needs.
     int deps_len = strlen(dep_libs);
     char * curr_dep = &dep_libs[0];
+
+    // We keep track of "special" libraries names (ones whose name is prefixed with `@`)
+    // which are libraries that we want to load in some special, custom way, such as
+    // `libjulia-internal` or `libjulia-codegen`.
+    int special_idx = 0;
+    char * special_library_names[2] = {NULL};
     while (1) {
-        // try to find next colon character, if we can't, escape out.
+        // try to find next colon character; if we can't, break out
         char * colon = strchr(curr_dep, ':');
         if (colon == NULL)
             break;
 
-        // Chop the string at the colon, load this library.
+        // Chop the string at the colon so it's a valid-ending-string
         *colon = '\0';
-        load_library(curr_dep, lib_dir);
+
+        // If this library name starts with `@`, don't open it here (but mark it as special)
+        if (curr_dep[0] == '@') {
+            if (special_idx > sizeof(special_library_names)/sizeof(char *)) {
+                jl_loader_print_stderr("ERROR: Too many special library names specified, check LOADER_BUILD_DEP_LIBS and friends!\n");
+                exit(1);
+            }
+            special_library_names[special_idx] = curr_dep + 1;
+            special_idx += 1;
+        } else {
+            load_library(curr_dep, lib_dir, 1);
+        }
 
         // Skip ahead to next dependency
         curr_dep = colon + 1;
     }
 
-    // Last dependency is `libjulia-internal`, so load that and we're done with `dep_libs`!
-    libjulia_internal = load_library(curr_dep, lib_dir);
+    if (special_idx != sizeof(special_library_names)/sizeof(char *)) {
+        jl_loader_print_stderr("ERROR: Too few special library names specified, check LOADER_BUILD_DEP_LIBS and friends!\n");
+        exit(1);
+    }
+
+    // Unpack our special library names.  This is why ordering of library names matters.
+    libjulia_internal = load_library(special_library_names[0], lib_dir, 1);
+    void *libjulia_codegen = load_library(special_library_names[1], lib_dir, 0);
+    const char * const * codegen_func_names;
+    const char *codegen_liberr;
+    if (libjulia_codegen == NULL) {
+        // if codegen is not available, use fallback implementation in libjulia-internal
+        libjulia_codegen = libjulia_internal;
+        codegen_func_names = jl_codegen_fallback_func_names;
+        codegen_liberr = " from libjulia-internal\n";
+    }
+    else {
+        codegen_func_names = jl_codegen_exported_func_names;
+        codegen_liberr = " from libjulia-codegen\n";
+    }
 
     // Once we have libjulia-internal loaded, re-export its symbols:
-    for (unsigned int symbol_idx=0; jl_exported_func_names[symbol_idx] != NULL; ++symbol_idx) {
-        void *addr = lookup_symbol(libjulia_internal, jl_exported_func_names[symbol_idx]);
-        if (addr == NULL || addr == *jl_exported_func_addrs[symbol_idx]) {
-            jl_loader_print_stderr3("ERROR: Unable to load ", jl_exported_func_names[symbol_idx], " from libjulia-internal");
+    for (unsigned int symbol_idx=0; jl_runtime_exported_func_names[symbol_idx] != NULL; ++symbol_idx) {
+        void *addr = lookup_symbol(libjulia_internal, jl_runtime_exported_func_names[symbol_idx]);
+        if (addr == NULL) {
+            jl_loader_print_stderr3("ERROR: Unable to load ", jl_runtime_exported_func_names[symbol_idx], " from libjulia-internal\n");
             exit(1);
         }
-        (*jl_exported_func_addrs[symbol_idx]) = addr;
+        (*jl_runtime_exported_func_addrs[symbol_idx]) = addr;
+    }
+    // jl_options must be initialized very early, in case an embedder sets some
+    // values there before calling jl_init
+    ((void (*)(void))jl_init_options_addr)();
+
+    for (unsigned int symbol_idx=0; codegen_func_names[symbol_idx] != NULL; ++symbol_idx) {
+        void *addr = lookup_symbol(libjulia_codegen, codegen_func_names[symbol_idx]);
+        if (addr == NULL) {
+            jl_loader_print_stderr3("ERROR: Unable to load ", codegen_func_names[symbol_idx], codegen_liberr);
+            exit(1);
+        }
+        (*jl_codegen_exported_func_addrs[symbol_idx]) = addr;
     }
     // Next, if we're on Linux/FreeBSD, set up fast TLS.
 #if !defined(_OS_WINDOWS_) && !defined(_OS_DARWIN_)
