@@ -79,6 +79,7 @@ JL_DLLEXPORT jl_typename_t *jl_new_typename_in(jl_sym_t *name, jl_module_t *modu
     tn->mt = NULL;
     tn->partial = NULL;
     tn->atomicfields = NULL;
+    tn->constfields = NULL;
     return tn;
 }
 
@@ -317,7 +318,7 @@ int jl_pointer_egal(jl_value_t *t)
         return 1;
     if (t == (jl_value_t*)jl_bool_type)
         return 1;
-    if (jl_is_mutable_datatype(t) && // excludes abstract types
+    if (jl_is_mutable_datatype(jl_unwrap_unionall(t)) && // excludes abstract types
         t != (jl_value_t*)jl_string_type && // technically mutable, but compared by contents
         t != (jl_value_t*)jl_simplevector_type &&
         !jl_is_kind(t))
@@ -340,6 +341,10 @@ int jl_pointer_egal(jl_value_t *t)
             return 1;
         }
     }
+    if (jl_is_uniontype(t)) {
+        jl_uniontype_t *u = (jl_uniontype_t*)t;
+        return jl_pointer_egal(u->a) && jl_pointer_egal(u->b);
+    }
     return 0;
 }
 
@@ -591,7 +596,7 @@ JL_DLLEXPORT jl_datatype_t *jl_new_datatype(
 
     t->name = NULL;
     if (jl_is_typename(name)) {
-        // This code-path is used by the Serialization module to by-pass normal expectations
+        // This code-path is used by the Serialization module to bypass normal expectations
         tn = (jl_typename_t*)name;
         tn->abstract = abstract;
         tn->mutabl = mutabl;
@@ -618,6 +623,7 @@ JL_DLLEXPORT jl_datatype_t *jl_new_datatype(
     tn->n_uninitialized = jl_svec_len(fnames) - ninitialized;
 
     uint32_t *volatile atomicfields = NULL;
+    uint32_t *volatile constfields = NULL;
     int i;
     JL_TRY {
         for (i = 0; i + 1 < jl_svec_len(fattrs); i += 2) {
@@ -629,7 +635,7 @@ JL_DLLEXPORT jl_datatype_t *jl_new_datatype(
             if (fldn < 1 || fldn > jl_svec_len(fnames))
                 jl_errorf("invalid field attribute %lld", (long long)fldn);
             fldn--;
-            if (attr == atomic_sym) {
+            if (attr == jl_atomic_sym) {
                 if (!mutabl)
                     jl_errorf("invalid field attribute atomic for immutable struct");
                 if (atomicfields == NULL) {
@@ -639,17 +645,28 @@ JL_DLLEXPORT jl_datatype_t *jl_new_datatype(
                 }
                 atomicfields[fldn / 32] |= 1 << (fldn % 32);
             }
+            else if (attr == jl_const_sym) {
+                if (!mutabl)
+                    jl_errorf("invalid field attribute const for immutable struct");
+                if (constfields == NULL) {
+                    size_t nb = (jl_svec_len(fnames) + 31) / 32 * sizeof(uint32_t);
+                    constfields = (uint32_t*)malloc_s(nb);
+                    memset(constfields, 0, nb);
+                }
+                constfields[fldn / 32] |= 1 << (fldn % 32);
+            }
             else {
                 jl_errorf("invalid field attribute %s", jl_symbol_name(attr));
             }
         }
     }
     JL_CATCH {
-        if (atomicfields)
-            free(atomicfields);
+        free(atomicfields);
+        free(constfields);
         jl_rethrow();
     }
     tn->atomicfields = atomicfields;
+    tn->constfields = constfields;
 
     if (t->name->wrapper == NULL) {
         t->name->wrapper = (jl_value_t*)t;
@@ -1713,7 +1730,7 @@ jl_value_t *replace_nth_field(jl_datatype_t *st, jl_value_t *v, size_t i, jl_val
         }
         else {
             jl_task_t *ct = jl_current_task;
-            uint8_t *psel;
+            uint8_t *psel = NULL;
             if (isunion) {
                 psel = &((uint8_t*)v)[offs + fsz - 1];
                 rty = jl_nth_union_component(rty, *psel);
@@ -1770,9 +1787,9 @@ JL_DLLEXPORT int jl_field_isdefined(jl_value_t *v, size_t i) JL_NOTSAFEPOINT
     return fval != NULL ? 1 : 0;
 }
 
-JL_DLLEXPORT size_t jl_get_field_offset(jl_datatype_t *ty, int field) JL_NOTSAFEPOINT
+JL_DLLEXPORT size_t jl_get_field_offset(jl_datatype_t *ty, int field)
 {
-    if (ty->layout == NULL || field > jl_datatype_nfields(ty) || field < 1)
+    if (!jl_struct_try_layout(ty) || field > jl_datatype_nfields(ty) || field < 1)
         jl_bounds_error_int((jl_value_t*)ty, field);
     return jl_field_offset(ty, field - 1);
 }
