@@ -4,13 +4,10 @@
 #include <errno.h>
 
 #include "julia.h"
+#include "julia_internal.h"
 
-#ifndef _MSC_VER
 #include <unistd.h>
 #include <getopt.h>
-#else
-#include "getopt.h"
-#endif
 #include "julia_assert.h"
 
 #ifdef _OS_WINDOWS_
@@ -27,7 +24,6 @@ JL_DLLEXPORT const char *jl_get_default_sysimg_path(void)
     return &system_image_path[1];
 }
 
-
 static int jl_options_initialized = 0;
 
 JL_DLLEXPORT void jl_init_options(void)
@@ -53,6 +49,7 @@ JL_DLLEXPORT void jl_init_options(void)
                         JL_OPTIONS_COMPILE_DEFAULT, // compile_enabled
                         0,    // code_coverage
                         0,    // malloc_log
+                        NULL, // tracked_path
                         2,    // opt_level
                         0,    // opt_level_min
 #ifdef JL_DEBUG_BUILD
@@ -78,105 +75,122 @@ JL_DLLEXPORT void jl_init_options(void)
                         NULL, // output-o
                         NULL, // output-asm
                         NULL, // output-ji
-                        NULL,    // output-code_coverage
+                        NULL, // output-code_coverage
                         0, // incremental
                         0, // image_file_specified
                         JL_OPTIONS_WARN_SCOPE_ON,  // ambiguous scope warning
                         0, // image-codegen
                         0, // rr-detach
+                        0, // strip-metadata
+                        0, // strip-ir
     };
     jl_options_initialized = 1;
 }
 
-static const char usage[] = "julia [switches] -- [programfile] [args...]\n";
+static const char usage[] = "\n    julia [switches] -- [programfile] [args...]\n\n";
 static const char opts[]  =
-    " -v, --version             Display version information\n"
-    " -h, --help                Print this message (--help-hidden for more)\n"
-    " --help-hidden             Uncommon options not shown by `-h`\n\n"
+    "Switches (a '*' marks the default value, if applicable):\n\n"
+    " -v, --version              Display version information\n"
+    " -h, --help                 Print this message (--help-hidden for more)\n"
+    " --help-hidden              Uncommon options not shown by `-h`\n\n"
 
     // startup options
-    " --project[={<dir>|@.}]    Set <dir> as the home project/environment\n"
-    " -J, --sysimage <file>     Start up with the given system image file\n"
-    " -H, --home <dir>          Set location of `julia` executable\n"
-    " --startup-file={yes|no}   Load `~/.julia/config/startup.jl`\n"
-    " --handle-signals={yes|no} Enable or disable Julia's default signal handlers\n"
-    " --sysimage-native-code={yes|no}\n"
-    "                           Use native code from system image if available\n"
-    " --compiled-modules={yes|no}\n"
-    "                           Enable or disable incremental precompilation of modules\n\n"
+    " --project[={<dir>|@.}]     Set <dir> as the home project/environment\n"
+    " -J, --sysimage <file>      Start up with the given system image file\n"
+    " -H, --home <dir>           Set location of `julia` executable\n"
+    " --startup-file={yes*|no}   Load `JULIA_DEPOT_PATH/config/startup.jl`; if `JULIA_DEPOT_PATH`\n"
+    "                            environment variable is unset, load `~/.julia/config/startup.jl`\n"
+    " --handle-signals={yes*|no} Enable or disable Julia's default signal handlers\n"
+    " --sysimage-native-code={yes*|no}\n"
+    "                            Use native code from system image if available\n"
+    " --compiled-modules={yes*|no}\n"
+    "                            Enable or disable incremental precompilation of modules\n\n"
 
     // actions
-    " -e, --eval <expr>         Evaluate <expr>\n"
-    " -E, --print <expr>        Evaluate <expr> and display the result\n"
-    " -L, --load <file>         Load <file> immediately on all processors\n\n"
+    " -e, --eval <expr>          Evaluate <expr>\n"
+    " -E, --print <expr>         Evaluate <expr> and display the result\n"
+    " -L, --load <file>          Load <file> immediately on all processors\n\n"
 
     // parallel options
-    " -t, --threads {N|auto}    Enable N threads; \"auto\" currently sets N to the number of local\n"
-    "                           CPU threads but this might change in the future\n"
-    " -p, --procs {N|auto}      Integer value N launches N additional local worker processes\n"
-    "                           \"auto\" launches as many workers as the number of local CPU threads (logical cores)\n"
-    " --machine-file <file>     Run processes on hosts listed in <file>\n\n"
+    " -t, --threads {N|auto}     Enable N threads; \"auto\" currently sets N to the number of local\n"
+    "                            CPU threads but this might change in the future\n"
+    " -p, --procs {N|auto}       Integer value N launches N additional local worker processes\n"
+    "                            \"auto\" launches as many workers as the number of local CPU threads (logical cores)\n"
+    " --machine-file <file>      Run processes on hosts listed in <file>\n\n"
 
     // interactive options
-    " -i                        Interactive mode; REPL runs and isinteractive() is true\n"
-    " -q, --quiet               Quiet startup: no banner, suppress REPL warnings\n"
-    " --banner={yes|no|auto}    Enable or disable startup banner\n"
-    " --color={yes|no|auto}     Enable or disable color text\n"
-    " --history-file={yes|no}   Load or save history\n\n"
+    " -i                         Interactive mode; REPL runs and `isinteractive()` is true\n"
+    " -q, --quiet                Quiet startup: no banner, suppress REPL warnings\n"
+    " --banner={yes|no|auto*}    Enable or disable startup banner\n"
+    " --color={yes|no|auto*}     Enable or disable color text\n"
+    " --history-file={yes*|no}   Load or save history\n\n"
 
     // error and warning options
-    " --depwarn={yes|no|error}  Enable or disable syntax and method deprecation warnings (\"error\" turns warnings into errors)\n"
-    " --warn-overwrite={yes|no} Enable or disable method overwrite warnings\n"
-    " --warn-scope={yes|no}     Enable or disable warning for ambiguous top-level scope\n\n"
+    " --depwarn={yes|no*|error}  Enable or disable syntax and method deprecation warnings (`error` turns warnings into errors)\n"
+    " --warn-overwrite={yes|no*} Enable or disable method overwrite warnings\n"
+    " --warn-scope={yes*|no}     Enable or disable warning for ambiguous top-level scope\n\n"
 
     // code generation options
-    " -C, --cpu-target <target> Limit usage of CPU features up to <target>; set to \"help\" to see the available options\n"
-    " -O, --optimize={0,1,2,3}  Set the optimization level (default level is 2 if unspecified or 3 if used without a level)\n"
-    " --min-optlevel={0,1,2,3}  Set a lower bound on the optimization level (default is 0)\n"
-    " -g, -g <level>            Enable / Set the level of debug info generation"
+    " -C, --cpu-target <target>  Limit usage of CPU features up to <target>; set to `help` to see the available options\n"
+    " -O, --optimize={0,1,2*,3}  Set the optimization level (level 3 if `-O` is used without a level)\n"
+    " --min-optlevel={0*,1,2,3}  Set a lower bound on the optimization level\n"
 #ifdef JL_DEBUG_BUILD
-        " (default level for julia-debug is 2 if unspecified or if used without a level)\n"
+        " -g [{0,1,2*}]              Set the level of debug info generation in the julia-debug build\n"
 #else
-        " (default level is 1 if unspecified or 2 if used without a level)\n"
+        " -g [{0,1*,2}]              Set the level of debug info generation (level 2 if `-g` is used without a level)\n"
 #endif
-    " --inline={yes|no}         Control whether inlining is permitted, including overriding @inline declarations\n"
-    " --check-bounds={yes|no|auto}\n"
-    "                           Emit bounds checks always, never, or respect @inbounds declarations\n"
+    " --inline={yes*|no}         Control whether inlining is permitted, including overriding @inline declarations\n"
+    " --check-bounds={yes|no|auto*}\n"
+    "                            Emit bounds checks always, never, or respect @inbounds declarations\n"
 #ifdef USE_POLLY
-    " --polly={yes|no}          Enable or disable the polyhedral optimizer Polly (overrides @polly declaration)\n"
+    " --polly={yes*|no}          Enable or disable the polyhedral optimizer Polly (overrides @polly declaration)\n"
 #endif
-    " --math-mode={ieee,fast}   Disallow or enable unsafe floating point optimizations (overrides @fastmath declaration)\n\n"
+    " --math-mode={ieee,fast}    Disallow or enable unsafe floating point optimizations (overrides @fastmath declaration)\n\n"
 
     // instrumentation options
-    " --code-coverage={none|user|all}, --code-coverage\n"
-    "                           Count executions of source lines (omitting setting is equivalent to \"user\")\n"
+    " --code-coverage[={none*|user|all}]\n"
+    "                            Count executions of source lines (omitting setting is equivalent to `user`)\n"
+    " --code-coverage=@<path>\n"
+    "                            Count executions but only in files that fall under the given file path/directory.\n"
+    "                            The `@` prefix is required to select this option. A `@` with no path will track the\n"
+    "                            current directory.\n"
+
     " --code-coverage=tracefile.info\n"
-    "                           Append coverage information to the LCOV tracefile (filename supports format tokens).\n"
+    "                            Append coverage information to the LCOV tracefile (filename supports format tokens)\n"
 // TODO: These TOKENS are defined in `runtime_ccall.cpp`. A more verbose `--help` should include that list here.
-    " --track-allocation={none|user|all}, --track-allocation\n"
-    "                           Count bytes allocated by each source line (omitting setting is equivalent to \"user\")\n"
-    " --bug-report=KIND         Launch a bug report session. It can be used to start a REPL, run a script, or evaluate\n"
-    "                           expressions. It first tries to use BugReporting.jl installed in current environment and\n"
-    "                           fallbacks to the latest compatible BugReporting.jl if not. For more information, see\n"
-    "                           --bug-report=help.\n\n"
+    " --track-allocation[={none*|user|all}]\n"
+    "                            Count bytes allocated by each source line (omitting setting is equivalent to `user`)\n"
+    " --track-allocation=@<path>\n"
+    "                            Count bytes but only in files that fall under the given file path/directory.\n"
+    "                            The `@` prefix is required to select this option. A `@` with no path will track the\n"
+    "                            current directory.\n"
+    " --bug-report=KIND          Launch a bug report session. It can be used to start a REPL, run a script, or evaluate\n"
+    "                            expressions. It first tries to use BugReporting.jl installed in current environment and\n"
+    "                            fallbacks to the latest compatible BugReporting.jl if not. For more information, see\n"
+    "                            --bug-report=help.\n\n"
 ;
 
 static const char opts_hidden[]  =
+    "Switches (a '*' marks the default value, if applicable):\n\n"
     // code generation options
-    " --compile={yes|no|all|min}Enable or disable JIT compiler, or request exhaustive or minimal compilation\n"
+    " --compile={yes*|no|all|min}\n"
+    "                          Enable or disable JIT compiler, or request exhaustive or minimal compilation\n\n"
 
     // compiler output options
-    " --output-o name           Generate an object file (including system image data)\n"
-    " --output-ji name          Generate a system image data file (.ji)\n"
+    " --output-o <name>        Generate an object file (including system image data)\n"
+    " --output-ji <name>       Generate a system image data file (.ji)\n"
+    " --strip-metadata         Remove docstrings and source location info from system image\n"
+    " --strip-ir               Remove IR (intermediate representation) of compiled functions\n\n"
 
     // compiler debugging (see the devdocs for tips on using these options)
-    " --output-unopt-bc name    Generate unoptimized LLVM bitcode (.bc)\n"
-    " --output-bc name          Generate LLVM bitcode (.bc)\n"
-    " --output-asm name         Generate an assembly file (.s)\n"
-    " --output-incremental=no   Generate an incremental output file (rather than complete)\n"
+    " --output-unopt-bc <name> Generate unoptimized LLVM bitcode (.bc)\n"
+    " --output-bc <name>       Generate LLVM bitcode (.bc)\n"
+    " --output-asm <name>      Generate an assembly file (.s)\n"
+    " --output-incremental={yes|no*}\n"
+    "                          Generate an incremental output file (rather than complete)\n"
     " --trace-compile={stderr,name}\n"
-    "                           Print precompile statements for methods compiled during execution or save to a path\n\n"
-    " --image-codegen           Force generate code in imaging mode\n"
+    "                          Print precompile statements for methods compiled during execution or save to a path\n"
+    " --image-codegen          Force generate code in imaging mode\n"
 ;
 
 JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
@@ -217,6 +231,8 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
            opt_bug_report,
            opt_image_codegen,
            opt_rr_detach,
+           opt_strip_metadata,
+           opt_strip_ir,
     };
     static const char* const shortopts = "+vhqH:e:E:L:J:C:it:p:O:g:";
     static const struct option longopts[] = {
@@ -235,7 +251,7 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
         { "bug-report",      required_argument, 0, opt_bug_report },
         { "sysimage",        required_argument, 0, 'J' },
         { "sysimage-native-code", required_argument, 0, opt_sysimage_native_code },
-        { "compiled-modules",    required_argument, 0, opt_compiled_modules },
+        { "compiled-modules",required_argument, 0, opt_compiled_modules },
         { "cpu-target",      required_argument, 0, 'C' },
         { "procs",           required_argument, 0, 'p' },
         { "threads",         required_argument, 0, 't' },
@@ -270,6 +286,8 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
         { "lisp",            no_argument,       0, 1 },
         { "image-codegen",   no_argument,       0, opt_image_codegen },
         { "rr-detach",       no_argument,       0, opt_rr_detach },
+        { "strip-metadata",  no_argument,       0, opt_strip_metadata },
+        { "strip-ir",        no_argument,       0, opt_strip_ir },
         { 0, 0, 0, 0 }
     };
 
@@ -507,6 +525,10 @@ restart_switch:
                         codecov = JL_LOG_ALL;
                     jl_options.output_code_coverage = optarg;
                 }
+                else if (!strncmp(optarg, "@", 1)) {
+                    codecov = JL_LOG_PATH;
+                    jl_options.tracked_path = optarg + 1; // skip `@`
+                }
                 else
                     jl_errorf("julia: invalid argument to --code-coverage (%s)", optarg);
                 break;
@@ -523,6 +545,10 @@ restart_switch:
                     malloclog = JL_LOG_ALL;
                 else if (!strcmp(optarg,"none"))
                     malloclog = JL_LOG_NONE;
+                else if (!strncmp(optarg, "@", 1)) {
+                    malloclog = JL_LOG_PATH;
+                    jl_options.tracked_path = optarg + 1; // skip `@`
+                }
                 else
                     jl_errorf("julia: invalid argument to --track-allocation (%s)", optarg);
                 break;
@@ -694,6 +720,12 @@ restart_switch:
         case opt_rr_detach:
             jl_options.rr_detach = 1;
             break;
+        case opt_strip_metadata:
+            jl_options.strip_metadata = 1;
+            break;
+        case opt_strip_ir:
+            jl_options.strip_ir = 1;
+            break;
         default:
             jl_errorf("julia: unhandled option -- %c\n"
                       "This is a bug, please report it.", c);
