@@ -36,13 +36,12 @@
 #include "julia_internal.h"
 #include "threading.h"
 #include "julia_assert.h"
-#include "support/hashing.h"
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-#if defined(JL_ASAN_ENABLED)
+#if defined(_COMPILER_ASAN_ENABLED_)
 static inline void sanitizer_start_switch_fiber(const void* bottom, size_t size) {
     __sanitizer_start_switch_fiber(NULL, bottom, size);
 }
@@ -54,16 +53,52 @@ static inline void sanitizer_start_switch_fiber(const void* bottom, size_t size)
 static inline void sanitizer_finish_switch_fiber(void) {}
 #endif
 
-#if defined(JL_TSAN_ENABLED)
-static inline void tsan_destroy_ctx(jl_ptls_t ptls, void *state) {
-    if (state != &ptls->root_task->state) {
-        __tsan_destroy_fiber(ctx->state);
-    }
-    ctx->state = NULL;
-}
-static inline void tsan_switch_to_ctx(void *state)  {
-    __tsan_switch_to_fiber(state, 0);
-}
+#if defined(_COMPILER_TSAN_ENABLED_)
+// must defined as macros, since the function containing them must not return before the longjmp
+#define tsan_destroy_ctx(_ptls, _ctx) do { \
+        jl_ucontext_t *_tsan_macro_ctx = (_ctx); \
+        if (_tsan_macro_ctx != &(_ptls)->root_task->ctx) { \
+            __tsan_destroy_fiber(_tsan_macro_ctx->tsan_state); \
+        } \
+        _tsan_macro_ctx->tsan_state = NULL; \
+    } while (0)
+#define tsan_switch_to_ctx(_ctx) do { \
+        jl_ucontext_t *_tsan_macro_ctx = (_ctx); \
+        __tsan_switch_to_fiber(_tsan_macro_ctx->tsan_state, 0); \
+    } while (0)
+#ifdef COPY_STACKS
+#define tsan_destroy_copyctx(_ptls, _ctx) do { \
+        jl_ucontext_t *_tsan_macro_ctx = (_ctx); \
+        if (_tsan_macro_ctx != &(_ptls)->root_task->ctx) { \
+            __tsan_destroy_fiber(_tsan_macro_ctx->tsan_state); \
+        } \
+        _tsan_macro_ctx->tsan_state = NULL; \
+    } while (0)
+#define tsan_switch_to_copyctx(_ctx) do { \
+        struct jl_stack_context_t *_tsan_macro_ctx = (_ctx); \
+        __tsan_switch_to_fiber(_tsan_macro_ctx->tsan_state, 0); \
+    } while (0)
+#endif
+#else
+// just do minimal type-checking on the arguments
+#define tsan_destroy_ctx(_ptls, _ctx) do { \
+        jl_ucontext_t *_tsan_macro_ctx = (_ctx); \
+        (void)_tsan_macro_ctx; \
+    } while (0)
+#define tsan_switch_to_ctx(_ctx) do { \
+        jl_ucontext_t *_tsan_macro_ctx = (_ctx); \
+        (void)_tsan_macro_ctx; \
+    } while (0)
+#ifdef COPY_STACKS
+#define tsan_destroy_copyctx(_ptls, _ctx) do { \
+        jl_ucontext_t *_tsan_macro_ctx = (_ctx); \
+        (void)_tsan_macro_ctx; \
+    } while (0)
+#define tsan_switch_to_copyctx(_ctx) do { \
+        jl_ucontext_t *_tsan_macro_ctx = (_ctx); \
+        (void)_tsan_macro_ctx; \
+    } while (0)
+#endif
 #endif
 
 // empirically, jl_finish_task needs about 64k stack space to infer/run
@@ -83,8 +118,7 @@ static inline void tsan_switch_to_ctx(void *state)  {
 #define STATIC_OR_JS static
 #endif
 
-extern size_t jl_page_size;
-static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT;
+static char *jl_alloc_fiber(_jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT;
 STATIC_OR_JS void jl_set_fiber(jl_ucontext_t *t);
 STATIC_OR_JS void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t);
 STATIC_OR_JS void jl_start_fiber_swap(jl_ucontext_t *savet, jl_ucontext_t *t);
@@ -151,9 +185,9 @@ static void NOINLINE JL_NORETURN restore_stack(jl_task_t *t, jl_ptls_t ptls, cha
 
     sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
 #if defined(_OS_WINDOWS_)
-    jl_setcontext(&t->ctx);
+    jl_setcontext(&t->ctx.copy_ctx);
 #else
-    jl_longjmp(t->copy_stack_ctx.uc_mcontext, 1);
+    jl_longjmp(t->ctx.copy_ctx.uc_mcontext, 1);
 #endif
     abort(); // unreachable
 }
@@ -168,22 +202,23 @@ static void restore_stack2(jl_task_t *t, jl_ptls_t ptls, jl_task_t *lastt)
     memcpy_a16((uint64_t*)_x, (uint64_t*)_y, nb); // destroys all but the current stackframe
 #if defined(JL_HAVE_UNW_CONTEXT)
     volatile int returns = 0;
-    int r = unw_getcontext(&lastt->ctx);
+    int r = unw_getcontext(&lastt->ctx.ctx);
     if (++returns == 2) // r is garbage after the first return
         return;
     if (r != 0 || returns != 1)
         abort();
 #elif defined(JL_HAVE_ASM) || defined(JL_HAVE_SIGALTSTACK) || defined(_OS_WINDOWS_)
-    if (jl_setjmp(lastt->copy_stack_ctx.uc_mcontext, 0))
+    if (jl_setjmp(lastt->ctx.copy_ctx.uc_mcontext, 0))
         return;
 #else
 #error COPY_STACKS is incompatible with this platform
 #endif
     sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
+    tsan_switch_to_copyctx(&t->ctx);
 #if defined(_OS_WINDOWS_)
-    jl_setcontext(&t->ctx);
+    jl_setcontext(&t->ctx.copy_ctx);
 #else
-    jl_longjmp(t->copy_stack_ctx.uc_mcontext, 1);
+    jl_longjmp(t->ctx.copy_ctx.uc_mcontext, 1);
 #endif
 }
 #endif
@@ -194,8 +229,9 @@ static _Atomic(jl_function_t*) task_done_hook_func JL_GLOBALLY_ROOTED = NULL;
 void JL_NORETURN jl_finish_task(jl_task_t *t)
 {
     jl_task_t *ct = jl_current_task;
+    JL_PROBE_RT_FINISH_TASK(ct);
     JL_SIGATOMIC_BEGIN();
-    if (t->_isexception)
+    if (jl_atomic_load_relaxed(&t->_isexception))
         jl_atomic_store_release(&t->_state, JL_TASK_STATE_FAILED);
     else
         jl_atomic_store_release(&t->_state, JL_TASK_STATE_DONE);
@@ -204,7 +240,7 @@ void JL_NORETURN jl_finish_task(jl_task_t *t)
     // ensure that state is cleared
     ct->ptls->in_finalizer = 0;
     ct->ptls->in_pure_callback = 0;
-    ct->world_age = jl_world_counter;
+    ct->world_age = jl_atomic_load_acquire(&jl_world_counter);
     // let the runtime know this task is dead and find a new task to run
     jl_function_t *done = jl_atomic_load_relaxed(&task_done_hook_func);
     if (done == NULL) {
@@ -221,7 +257,7 @@ void JL_NORETURN jl_finish_task(jl_task_t *t)
             jl_no_exc_handler(jl_current_exception());
         }
     }
-    gc_debug_critical_error();
+    jl_gc_debug_critical_error();
     abort();
 }
 
@@ -240,7 +276,7 @@ JL_DLLEXPORT void *jl_task_stack_buffer(jl_task_t *task, size_t *size, int *ptid
     jl_ptls_t ptls2 = task->ptls;
     *ptid = -1;
     if (ptls2) {
-        *ptid = task->tid;
+        *ptid = jl_atomic_load_relaxed(&task->tid);
 #ifdef COPY_STACKS
         if (task->copy_stack) {
             *size = ptls2->stacksize;
@@ -322,7 +358,7 @@ JL_DLLEXPORT jl_task_t *jl_get_next_task(void) JL_NOTSAFEPOINT
     return ct;
 }
 
-#ifdef JL_TSAN_ENABLED
+#ifdef _COMPILER_TSAN_ENABLED_
 const char tsan_state_corruption[] = "TSAN state corrupted. Exiting HARD!\n";
 #endif
 
@@ -335,7 +371,7 @@ static void ctx_switch(jl_task_t *lastt)
     // none of these locks should be held across a task switch
     assert(ptls->locks.len == 0);
 
-#ifdef JL_TSAN_ENABLED
+#ifdef _COMPILER_TSAN_ENABLED_
     if (lastt->ctx.tsan_state != __tsan_get_current_fiber()) {
         // Something went really wrong - don't even assume that we can
         // use assert/abort which involve lots of signal handling that
@@ -345,11 +381,11 @@ static void ctx_switch(jl_task_t *lastt)
     }
 #endif
 
-    int killed = lastt->_state != JL_TASK_STATE_RUNNABLE;
+    int killed = jl_atomic_load_relaxed(&lastt->_state) != JL_TASK_STATE_RUNNABLE;
     if (!t->started && !t->copy_stack) {
         // may need to allocate the stack
         if (t->stkbuf == NULL) {
-            t->stkbuf = jl_alloc_fiber(&t->ctx, &t->bufsz, t);
+            t->stkbuf = jl_alloc_fiber(&t->ctx.ctx, &t->bufsz, t);
             if (t->stkbuf == NULL) {
 #ifdef COPY_STACKS
                 // fall back to stack copying if mmap fails
@@ -357,9 +393,9 @@ static void ctx_switch(jl_task_t *lastt)
                 t->sticky = 1;
                 t->bufsz = 0;
                 if (always_copy_stacks)
-                    memcpy(&t->copy_stack_ctx, &ptls->copy_stack_ctx, sizeof(t->copy_stack_ctx));
+                    memcpy(&t->ctx.copy_ctx, &ptls->copy_stack_ctx, sizeof(t->ctx.copy_ctx));
                 else
-                    memcpy(&t->ctx, &ptls->base_ctx, sizeof(t->ctx));
+                    memcpy(&t->ctx.ctx, &ptls->base_ctx, sizeof(t->ctx.ctx));
 #else
                 jl_throw(jl_memory_exception);
 #endif
@@ -379,7 +415,7 @@ static void ctx_switch(jl_task_t *lastt)
 #ifdef COPY_STACKS
         if (lastt->copy_stack) { // save the old copy-stack
             save_stack(ptls, lastt, pt); // allocates (gc-safepoint, and can also fail)
-            if (jl_setjmp(lastt->copy_stack_ctx.uc_mcontext, 0)) {
+            if (jl_setjmp(lastt->ctx.copy_ctx.uc_mcontext, 0)) {
                 sanitizer_finish_switch_fiber();
                 // TODO: mutex unlock the thread we just switched from
                 return;
@@ -392,7 +428,7 @@ static void ctx_switch(jl_task_t *lastt)
 
     // set up global state for new task and clear global state for old task
     t->ptls = ptls;
-    ptls->current_task = t;
+    jl_atomic_store_relaxed(&ptls->current_task, t);
     JL_GC_PROMISE_ROOTED(t);
     jl_signal_fence();
     jl_set_pgcstack(&t->gcstack);
@@ -402,21 +438,22 @@ static void ctx_switch(jl_task_t *lastt)
     ptls->previous_task = lastt;
 #endif
 
-#if defined(JL_TSAN_ENABLED)
-    tsan_switch_to_ctx(&t->tsan_state);
-    if (killed)
-        tsan_destroy_ctx(ptls, &lastt->tsan_state);
-#endif
     if (t->started) {
 #ifdef COPY_STACKS
         if (t->copy_stack) {
             if (!killed && !lastt->copy_stack)
                 restore_stack2(t, ptls, lastt);
-            else if (lastt->copy_stack) {
-                restore_stack(t, ptls, NULL); // (doesn't return)
-            }
             else {
-                restore_stack(t, ptls, (char*)1); // (doesn't return)
+                tsan_switch_to_copyctx(&t->ctx);
+                if (killed)
+                    tsan_destroy_copyctx(ptls, &lastt->ctx);
+
+                if (lastt->copy_stack) {
+                    restore_stack(t, ptls, NULL); // (doesn't return)
+                }
+                else {
+                    restore_stack(t, ptls, (char*)1); // (doesn't return)
+                }
             }
         }
         else
@@ -424,6 +461,8 @@ static void ctx_switch(jl_task_t *lastt)
         {
             sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
             if (killed) {
+                tsan_switch_to_ctx(&t->ctx);
+                tsan_destroy_ctx(ptls, &lastt->ctx);
                 jl_set_fiber(&t->ctx); // (doesn't return)
                 abort(); // unreachable
             }
@@ -431,6 +470,7 @@ static void ctx_switch(jl_task_t *lastt)
                 if (lastt->copy_stack) {
                     // Resume at the jl_setjmp earlier in this function,
                     // don't do a full task swap
+                    tsan_switch_to_ctx(&t->ctx);
                     jl_set_fiber(&t->ctx); // (doesn't return)
                 }
                 else {
@@ -442,22 +482,29 @@ static void ctx_switch(jl_task_t *lastt)
     else {
         sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
         if (t->copy_stack && always_copy_stacks) {
+            tsan_switch_to_ctx(&t->ctx);
+            if (killed) {
+                tsan_destroy_ctx(ptls, &lastt->ctx);
+            }
 #ifdef COPY_STACKS
 #if defined(_OS_WINDOWS_)
-            jl_setcontext(&t->ctx);
+            jl_setcontext(&t->ctx.copy_ctx);
 #else
-            jl_longjmp(t->copy_stack_ctx.uc_mcontext, 1);
+            jl_longjmp(t->ctx.copy_ctx.uc_mcontext, 1);
 #endif
 #endif
             abort(); // unreachable
         }
         else {
             if (killed) {
+                tsan_switch_to_ctx(&t->ctx);
+                tsan_destroy_ctx(ptls, &lastt->ctx);
                 jl_start_fiber_set(&t->ctx); // (doesn't return)
                 abort();
             }
             else if (lastt->copy_stack) {
                 // Resume at the jl_setjmp earlier in this function
+                tsan_switch_to_ctx(&t->ctx);
                 jl_start_fiber_set(&t->ctx); // (doesn't return)
                 abort();
             }
@@ -483,9 +530,11 @@ JL_DLLEXPORT void jl_switch(void)
         jl_error("task switch not allowed from inside gc finalizer");
     if (ptls->in_pure_callback)
         jl_error("task switch not allowed from inside staged nor pure functions");
-    if (!jl_set_task_tid(t, ptls->tid)) // manually yielding to a task
+    if (!jl_set_task_tid(t, jl_atomic_load_relaxed(&ct->tid))) // manually yielding to a task
         jl_error("cannot switch to task running on another thread");
 
+    JL_PROBE_RT_PAUSE_TASK(ct);
+
     // Store old values on the stack and reset
     sig_atomic_t defer_signal = ptls->defer_signal;
     int8_t gc_state = jl_gc_unsafe_enter(ptls);
@@ -506,7 +555,7 @@ JL_DLLEXPORT void jl_switch(void)
     t = ptls->previous_task;
     ptls->previous_task = NULL;
     assert(t != ct);
-    assert(t->tid == ptls->tid);
+    assert(jl_atomic_load_relaxed(&t->tid) == ptls->tid);
     if (!t->sticky && !t->copy_stack)
         jl_atomic_store_release(&t->tid, -1);
 #else
@@ -533,6 +582,8 @@ JL_DLLEXPORT void jl_switch(void)
     ptls->defer_signal = defer_signal;
     if (other_defer_signal && !defer_signal)
         jl_sigint_safepoint(ptls);
+
+    JL_PROBE_RT_RUN_TASK(ct);
 }
 
 JL_DLLEXPORT void jl_switchto(jl_task_t **pt)
@@ -644,21 +695,23 @@ JL_DLLEXPORT void jl_rethrow_other(jl_value_t *e JL_MAYBE_UNROOTED)
     throw_internal(ct, NULL);
 }
 
-/* This is xoshiro256++ 1.0, used for tasklocal random number generation in julia.
+/* This is xoshiro256++ 1.0, used for tasklocal random number generation in Julia.
    This implementation is intended for embedders and internal use by the runtime, and is
-   based on the reference implementation on http://prng.di.unimi.it
+   based on the reference implementation at https://prng.di.unimi.it
 
-   Credits go to Sebastiano Vigna for coming up with this PRNG.
+   Credits go to David Blackman and Sebastiano Vigna for coming up with this PRNG.
+   They described xoshiro256++ in "Scrambled Linear Pseudorandom Number Generators",
+   ACM Trans. Math. Softw., 2021.
 
-   There is a pure julia implementation in stdlib that tends to be faster when used from
-   within julia, due to inlining and more agressive architecture-specific optimizations.
+   There is a pure Julia implementation in stdlib that tends to be faster when used from
+   within Julia, due to inlining and more agressive architecture-specific optimizations.
 */
-JL_DLLEXPORT uint64_t jl_tasklocal_genrandom(jl_task_t *task) JL_NOTSAFEPOINT
+uint64_t jl_genrandom(uint64_t rngState[4]) JL_NOTSAFEPOINT
 {
-    uint64_t s0 = task->rngState0;
-    uint64_t s1 = task->rngState1;
-    uint64_t s2 = task->rngState2;
-    uint64_t s3 = task->rngState3;
+    uint64_t s0 = rngState[0];
+    uint64_t s1 = rngState[1];
+    uint64_t s2 = rngState[2];
+    uint64_t s3 = rngState[3];
 
     uint64_t t = s1 << 17;
     uint64_t tmp = s0 + s3;
@@ -670,14 +723,14 @@ JL_DLLEXPORT uint64_t jl_tasklocal_genrandom(jl_task_t *task) JL_NOTSAFEPOINT
     s2 ^= t;
     s3 = (s3 << 45) | (s3 >> 19);
 
-    task->rngState0 = s0;
-    task->rngState1 = s1;
-    task->rngState2 = s2;
-    task->rngState3 = s3;
+    rngState[0] = s0;
+    rngState[1] = s1;
+    rngState[2] = s2;
+    rngState[3] = s3;
     return res;
 }
 
-void rng_split(jl_task_t *from, jl_task_t *to) JL_NOTSAFEPOINT
+void jl_rng_split(uint64_t to[4], uint64_t from[4]) JL_NOTSAFEPOINT
 {
     /* TODO: consider a less ad-hoc construction
        Ideally we could just use the output of the random stream to seed the initial
@@ -695,16 +748,17 @@ void rng_split(jl_task_t *from, jl_task_t *to) JL_NOTSAFEPOINT
        0x3688cf5d48899fa7 == hash(UInt(3))|0x01
        0x867b4bb4c42e5661 == hash(UInt(4))|0x01
     */
-    to->rngState0 = 0x02011ce34bce797f * jl_tasklocal_genrandom(from);
-    to->rngState1 = 0x5a94851fb48a6e05 * jl_tasklocal_genrandom(from);
-    to->rngState2 = 0x3688cf5d48899fa7 * jl_tasklocal_genrandom(from);
-    to->rngState3 = 0x867b4bb4c42e5661 * jl_tasklocal_genrandom(from);
+    to[0] = 0x02011ce34bce797f * jl_genrandom(from);
+    to[1] = 0x5a94851fb48a6e05 * jl_genrandom(from);
+    to[2] = 0x3688cf5d48899fa7 * jl_genrandom(from);
+    to[3] = 0x867b4bb4c42e5661 * jl_genrandom(from);
 }
 
 JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion_future, size_t ssize)
 {
     jl_task_t *ct = jl_current_task;
     jl_task_t *t = (jl_task_t*)jl_gc_alloc(ct->ptls, sizeof(jl_task_t), jl_task_type);
+    JL_PROBE_RT_NEW_TASK(ct, t);
     t->copy_stack = 0;
     if (ssize == 0) {
         // stack size unspecified; use default
@@ -722,7 +776,7 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion
         if (ssize < MINSTKSZ)
             ssize = MINSTKSZ;
         t->bufsz = ssize;
-        t->stkbuf = jl_alloc_fiber(&t->ctx, &t->bufsz, t);
+        t->stkbuf = jl_alloc_fiber(&t->ctx.ctx, &t->bufsz, t);
         if (t->stkbuf == NULL)
             jl_throw(jl_memory_exception);
     }
@@ -733,11 +787,11 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion
     t->start = start;
     t->result = jl_nothing;
     t->donenotify = completion_future;
-    t->_isexception = 0;
+    jl_atomic_store_relaxed(&t->_isexception, 0);
     // Inherit logger state from parent task
     t->logstate = ct->logstate;
     // Fork task-local random state from parent
-    rng_split(ct, t);
+    jl_rng_split(t->rngState, ct->rngState);
     // there is no active exception handler available on this stack yet
     t->eh = NULL;
     t->sticky = 1;
@@ -745,9 +799,9 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion
     t->excstack = NULL;
     t->started = 0;
     t->prio = -1;
-    jl_atomic_store_relaxed(&t->tid, t->copy_stack ? ct->tid : -1); // copy_stacks are always pinned since they can't be moved
+    jl_atomic_store_relaxed(&t->tid, t->copy_stack ? jl_atomic_load_relaxed(&ct->tid) : -1); // copy_stacks are always pinned since they can't be moved
     t->ptls = NULL;
-    t->world_age = 0;
+    t->world_age = ct->world_age;
 
 #ifdef COPY_STACKS
     if (!t->copy_stack) {
@@ -757,13 +811,13 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion
     }
     else {
         if (always_copy_stacks)
-            memcpy(&t->copy_stack_ctx, &ct->ptls->copy_stack_ctx, sizeof(t->copy_stack_ctx));
+            memcpy(&t->ctx.copy_ctx, &ct->ptls->copy_stack_ctx, sizeof(t->ctx.copy_ctx));
         else
-            memcpy(&t->ctx, &ct->ptls->base_ctx, sizeof(t->ctx));
+            memcpy(&t->ctx.ctx, &ct->ptls->base_ctx, sizeof(t->ctx.ctx));
     }
 #endif
-#ifdef JL_TSAN_ENABLED
-    t->tsan_state = __tsan_create_fiber(0);
+#ifdef _COMPILER_TSAN_ENABLED_
+    t->ctx.tsan_state = __tsan_create_fiber(0);
 #endif
     return t;
 }
@@ -778,7 +832,7 @@ JL_DLLEXPORT jl_task_t *jl_get_current_task(void)
 #ifdef JL_HAVE_ASYNCIFY
 JL_DLLEXPORT jl_ucontext_t *task_ctx_ptr(jl_task_t *t)
 {
-    return &t->ctx;
+    return &t->ctx.ctx;
 }
 
 JL_DLLEXPORT jl_value_t *jl_get_root_task(void)
@@ -860,7 +914,8 @@ CFI_NORETURN
 #endif
 
     ct->started = 1;
-    if (ct->_isexception) {
+    JL_PROBE_RT_START_TASK(ct);
+    if (jl_atomic_load_relaxed(&ct->_isexception)) {
         record_backtrace(ptls, 0);
         jl_push_excstack(&ct->excstack, ct->result,
                          ptls->bt_data, ptls->bt_size);
@@ -873,12 +928,11 @@ CFI_NORETURN
                 jl_sigint_safepoint(ptls);
             }
             JL_TIMING(ROOT);
-            ct->world_age = jl_world_counter;
             res = jl_apply(&ct->start, 1);
         }
         JL_CATCH {
             res = jl_current_exception();
-            ct->_isexception = 1;
+            jl_atomic_store_relaxed(&ct->_isexception, 1);
             goto skip_pop_exception;
         }
 skip_pop_exception:;
@@ -886,7 +940,7 @@ skip_pop_exception:;
     ct->result = res;
     jl_gc_wb(ct, ct->result);
     jl_finish_task(ct);
-    gc_debug_critical_error();
+    jl_gc_debug_critical_error();
     abort();
 }
 
@@ -897,7 +951,7 @@ skip_pop_exception:;
 #define swapcontext jl_swapcontext
 #define makecontext jl_makecontext
 #endif
-static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT
+static char *jl_alloc_fiber(_jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT
 {
 #ifndef _OS_WINDOWS_
     int r = getcontext(t);
@@ -919,25 +973,27 @@ static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) J
 }
 static void jl_start_fiber_set(jl_ucontext_t *t)
 {
-    setcontext(t);
+    setcontext(&t->ctx);
 }
 static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
     assert(lastt);
-    swapcontext(lastt, t);
+    tsan_switch_to_ctx(t);
+    swapcontext(&lastt->ctx, &t->ctx);
 }
 static void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
-    swapcontext(lastt, t);
+    tsan_switch_to_ctx(t);
+    swapcontext(&lastt->ctx, &t->ctx);
 }
 static void jl_set_fiber(jl_ucontext_t *t)
 {
-    setcontext(t);
+    setcontext(&t->ctx);
 }
 #endif
 
 #if defined(JL_HAVE_UNW_CONTEXT) || defined(JL_HAVE_ASM)
-static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
+static char *jl_alloc_fiber(_jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
 {
     char *stkbuf = (char*)jl_malloc_stack(ssize, owner);
     if (stkbuf == NULL)
@@ -964,15 +1020,15 @@ static inline void jl_unw_swapcontext(unw_context_t *old, unw_cursor_t *c)
 static void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
     unw_cursor_t c;
-    int r = unw_init_local(&c, t);
+    int r = unw_init_local(&c, &t->ctx);
     if (r < 0)
         abort();
-    jl_unw_swapcontext(lastt, &c);
+    jl_unw_swapcontext(&lastt->ctx, &c);
 }
-static void jl_set_fiber(unw_context_t *t)
+static void jl_set_fiber(jl_ucontext_t *t)
 {
     unw_cursor_t c;
-    int r = unw_init_local(&c, t);
+    int r = unw_init_local(&c, &t->ctx);
     if (r < 0)
         abort();
     unw_resume(&c);
@@ -980,13 +1036,14 @@ static void jl_set_fiber(unw_context_t *t)
 #elif defined(JL_HAVE_ASM)
 static void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
-    if (jl_setjmp(lastt->uc_mcontext, 0))
+    if (jl_setjmp(lastt->ctx.uc_mcontext, 0))
         return;
+    tsan_switch_to_ctx(t);
     jl_set_fiber(t); // doesn't return
 }
 static void jl_set_fiber(jl_ucontext_t *t)
 {
-    jl_longjmp(t->uc_mcontext, 1);
+    jl_longjmp(t->ctx.uc_mcontext, 1);
 }
 #endif
 
@@ -1004,21 +1061,21 @@ static void jl_set_fiber(jl_ucontext_t *t)
 #else
 #error please define how to simulate a CALL on this platform
 #endif
-static void jl_start_fiber_set(unw_context_t *t)
+static void jl_start_fiber_set(jl_ucontext_t *t)
 {
     unw_cursor_t c;
-    char *stk = ((char**)t)[0];
-    size_t ssize = ((size_t*)t)[1];
+    char *stk = ((char**)&t->ctx)[0];
+    size_t ssize = ((size_t*)&t->ctx)[1];
     uintptr_t fn = (uintptr_t)&start_task;
     stk += ssize;
-    int r = unw_getcontext(t);
+    int r = unw_getcontext(&t->ctx);
     if (r)
         abort();
-    if (unw_init_local(&c, t))
+    if (unw_init_local(&c, &t->ctx))
         abort();
     PUSH_RET(&c, stk);
 #if defined __linux__
-#error savannah nongnu libunwind is not capable of setting UNW_REG_SP, as required
+#error savannah nongnu libunwind is incapable of setting UNW_REG_SP, as required
 #endif
     if (unw_set_reg(&c, UNW_REG_SP, (uintptr_t)stk))
         abort();
@@ -1026,31 +1083,31 @@ static void jl_start_fiber_set(unw_context_t *t)
         abort();
     unw_resume(&c); // (doesn't return)
 }
-static void jl_start_fiber_swap(unw_context_t *lastt, unw_context_t *t)
+static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
     assert(lastt);
     unw_cursor_t c;
-    char *stk = ((char**)t)[0];
-    size_t ssize = ((size_t*)t)[1];
+    char *stk = ((char**)&t->ctx)[0];
+    size_t ssize = ((size_t*)&t->ctx)[1];
     uintptr_t fn = (uintptr_t)&start_task;
     stk += ssize;
     volatile int returns = 0;
-    int r = unw_getcontext(lastt);
+    int r = unw_getcontext(&lastt->ctx);
     if (++returns == 2) // r is garbage after the first return
         return;
     if (r != 0 || returns != 1)
         abort();
-    int r = unw_getcontext(t);
+    r = unw_getcontext(&t->ctx);
     if (r != 0)
         abort();
-    if (unw_init_local(&c, t))
+    if (unw_init_local(&c, &t->ctx))
         abort();
     PUSH_RET(&c, stk);
     if (unw_set_reg(&c, UNW_REG_SP, (uintptr_t)stk))
         abort();
     if (unw_set_reg(&c, UNW_REG_IP, fn))
         abort();
-    jl_unw_swapcontext(lastt, &c);
+    jl_unw_swapcontext(&lastt->ctx, &c);
 }
 #endif
 
@@ -1060,21 +1117,22 @@ static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
     assert(lastt);
 #ifdef JL_HAVE_UNW_CONTEXT
     volatile int returns = 0;
-    int r = unw_getcontext(lastt);
+    int r = unw_getcontext(&lastt->ctx);
     if (++returns == 2) // r is garbage after the first return
         return;
     if (r != 0 || returns != 1)
         abort();
 #else
-    if (jl_setjmp(lastt->uc_mcontext, 0))
+    if (jl_setjmp(lastt->ctx.uc_mcontext, 0))
         return;
 #endif
+    tsan_switch_to_ctx(t);
     jl_start_fiber_set(t); // doesn't return
 }
 static void jl_start_fiber_set(jl_ucontext_t *t)
 {
-    char *stk = ((char**)t)[0];
-    size_t ssize = ((size_t*)t)[1];
+    char *stk = ((char**)&t->ctx)[0];
+    size_t ssize = ((size_t*)&t->ctx)[1];
     uintptr_t fn = (uintptr_t)&start_task;
     stk += ssize;
 #ifdef _CPU_X86_64_
@@ -1144,7 +1202,7 @@ static void jl_start_fiber_set(jl_ucontext_t *t)
 #endif
 
 #if defined(JL_HAVE_SIGALTSTACK)
-#if defined(JL_TSAN_ENABLED)
+#if defined(_COMPILER_TSAN_ENABLED_)
 #error TSAN support not currently implemented for this tasking model
 #endif
 
@@ -1154,7 +1212,7 @@ static void start_basefiber(int sig)
     if (jl_setjmp(ptls->base_ctx.uc_mcontext, 0))
         start_task(); // sanitizer_finish_switch_fiber is part of start_task
 }
-static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
+static char *jl_alloc_fiber(_jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
 {
     stack_t uc_stack, osigstk;
     struct sigaction sa, osa;
@@ -1164,8 +1222,8 @@ static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
         return NULL;
     // setup
     jl_ptls_t ptls = jl_current_task->ptls;
-    jl_ucontext_t base_ctx;
-    memcpy(&base_ctx, &ptls->base_ctx, sizeof(ptls->base_ctx));
+    _jl_ucontext_t base_ctx;
+    memcpy(&base_ctx, &ptls->base_ctx, sizeof(base_ctx));
     sigfillset(&set);
     if (sigprocmask(SIG_BLOCK, &set, &oset) != 0) {
        jl_free_stack(stk, *ssize);
@@ -1206,39 +1264,41 @@ static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
        jl_error("sigprocmask failed");
     }
     if (&ptls->base_ctx != t) {
-        memcpy(t, &ptls->base_ctx, sizeof(ptls->base_ctx));
-        memcpy(&ptls->base_ctx, &base_ctx, sizeof(ptls->base_ctx)); // restore COPY_STACKS context
+        memcpy(&t, &ptls->base_ctx, sizeof(base_ctx));
+        memcpy(&ptls->base_ctx, &base_ctx, sizeof(base_ctx)); // restore COPY_STACKS context
     }
     return (char*)stk;
 }
 static void jl_start_fiber_set(jl_ucontext_t *t) {
-    jl_longjmp(t->uc_mcontext, 1); // (doesn't return)
+    jl_longjmp(t->ctx.uc_mcontext, 1); // (doesn't return)
 }
 static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
     assert(lastt);
-    if (lastt && jl_setjmp(lastt->uc_mcontext, 0))
+    if (lastt && jl_setjmp(lastt->ctx.uc_mcontext, 0))
         return;
+    tsan_switch_to_ctx(t);
     jl_start_fiber_set(t);
 }
 static void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
-    if (jl_setjmp(lastt->uc_mcontext, 0))
+    if (jl_setjmp(lastt->ctx.uc_mcontext, 0))
         return;
+    tsan_switch_to_ctx(t);
     jl_start_fiber_set(t); // doesn't return
 }
 static void jl_set_fiber(jl_ucontext_t *t)
 {
-    jl_longjmp(t->uc_mcontext, 1);
+    jl_longjmp(t->ctx.uc_mcontext, 1);
 }
 #endif
 
 #if defined(JL_HAVE_ASYNCIFY)
-#if defined(JL_TSAN_ENABLED)
+#if defined(_COMPILER_TSAN_ENABLED_)
 #error TSAN support not currently implemented for this tasking model
 #endif
 
-static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT
+static char *jl_alloc_fiber(_jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT
 {
     void *stk = jl_malloc_stack(ssize, owner);
     if (stk == NULL)
@@ -1295,7 +1355,7 @@ jl_task_t *jl_init_root_task(jl_ptls_t ptls, void *stack_lo, void *stack_hi)
     ct->start = NULL;
     ct->result = jl_nothing;
     ct->donenotify = jl_nothing;
-    ct->_isexception = 0;
+    jl_atomic_store_relaxed(&ct->_isexception, 0);
     ct->logstate = jl_nothing;
     ct->eh = NULL;
     ct->gcstack = NULL;
@@ -1305,13 +1365,13 @@ jl_task_t *jl_init_root_task(jl_ptls_t ptls, void *stack_lo, void *stack_hi)
     ct->ptls = ptls;
     ct->world_age = 1; // OK to run Julia code on this task
     ptls->root_task = ct;
-    ptls->current_task = ct;
+    jl_atomic_store_relaxed(&ptls->current_task, ct);
     JL_GC_PROMISE_ROOTED(ct);
     jl_set_pgcstack(&ct->gcstack);
     assert(jl_current_task == ct);
 
-#ifdef JL_TSAN_ENABLED
-    ct->tsan_state = __tsan_get_current_fiber();
+#ifdef _COMPILER_TSAN_ENABLED_
+    ct->ctx.tsan_state = __tsan_get_current_fiber();
 #endif
 
 #ifdef COPY_STACKS
@@ -1344,7 +1404,7 @@ JL_DLLEXPORT int jl_is_task_started(jl_task_t *t) JL_NOTSAFEPOINT
 
 JL_DLLEXPORT int16_t jl_get_task_tid(jl_task_t *t) JL_NOTSAFEPOINT
 {
-    return t->tid;
+    return jl_atomic_load_relaxed(&t->tid);
 }
 
 
